from dataclasses import dataclass, field
from typing import List


@dataclass
class CharInspection:

    """
    Schema representing an inspection of an attempt word char.
    """

    char: str
    position: int
    valid: bool = False
    missplaced: bool = False


@dataclass
class Attempt:

    """
    Schema representing a database `Attempt` record.
    """

    id: int
    word: str
    match: bool
    chars: List[CharInspection] = field(default_factory=list)


@dataclass
class Game:

    """
    Schema representing a database `Game` record.
    """

    id: int
    word: str
    date: str
    attempts: List[Attempt] = field(default_factory=list)
    max_attempts: int = 6

    def any_match(self) -> bool:

        """
        Check if any of the attempts did a match.

        Returns:
            True if a match is found; otherwise False.
        """

        return any([attempt.match for attempt in self.attempts])

    def any_attempt_left(self) -> bool:

        """
        Check if there is any attempt left.

        Returns:
            True when one or more attempts are left; otherwise, False.
        """

        return len(self.attempts) < self.max_attempts

    def get_attempts_left(self) -> int:

        """
        Get the number of the attempts left.

        Returns:
            The number of the attempts left.
        """

        return self.max_attempts - len(self.attempts)



import json
import os
import sqlite3
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from palabros import __version__
from palabros.errors import DatabaseError
from palabros.queries import Query

_APP_PATH = Path.home().joinpath(".palabros")
_DB_PATH = _APP_PATH.joinpath("games.db")
_WORDS_JSON_PATH = Path(__file__).parent.joinpath("words.json")


@contextmanager
def db_cursor() -> Generator[sqlite3.Cursor, None, None]:

    """
    Create a cursor to interact with the database.

    Yields:
        An instance of `sqlite3.Cursor`.

    Raises:
        DatabaseError: if any error occurred while working with the database.
    """

    try:
        with sqlite3.connect(
            os.getenv("PALABROS_DB_PATH", str(_DB_PATH)),
            isolation_level=None,
        ) as conn:
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()
            yield cur
    except sqlite3.DatabaseError as err:
        raise DatabaseError(str(err)) from err


def _generate_words_population_query() -> str:

    """
    Generate the query to populate the words table.

    Returns:
        The query, ready to execute.
    """

    with open(_WORDS_JSON_PATH) as f:
        words = json.load(f)
    values = ", ".join([f"('{k}', {v})" for k, v in words.items()])
    return Query.PopulateWordsTable.format(values)


def _table_exists(table: str) -> bool:

    """
    Check if the given table exists in the database.

    Args:
        table: The table to check.

    Returns:
        True if the table exists in the database; otherwise False.
    """

    with db_cursor() as cur:
        cur.execute(Query.TableExists, (table,))
        result = cur.fetchone()

    return result is not None and bool(result["total"])


def init() -> None:

    """
    Initialize the database.
    """

    if not _APP_PATH.exists():
        _APP_PATH.mkdir(parents=True, exist_ok=True)

    if _DB_PATH.exists():
        if _table_exists("versions"):
            return
        # Force old database rebuilding
        _DB_PATH.unlink(missing_ok=True)

    with db_cursor() as cur:
        cur.execute(Query.CreateVersionsTable)
        cur.execute(Query.CreateWordsTable)
        cur.execute(Query.CreateGamesTable)
        cur.execute(Query.CreateAttemptsTable)
        cur.execute(Query.CreateGamesDateIndex)
        cur.execute(Query.CreateVersion, (__version__,))
        cur.execute(_generate_words_population_query())



from collections import Counter
from datetime import datetime
from typing import List

from palabros.database import db_cursor
from palabros.errors import DatabaseError, GameError
from palabros.queries import Query
from palabros.schemas import Attempt, CharInspection, Game

WORD_LENGTH = 5


def _normalize_word(word: str) -> str:

    """
    Normalize a given word.

    Args:
        word: The word to normalize.

    Returns:
        The properly normalized word.
    """

    return word.translate(str.maketrans("áéíóúü", "aeiouu"))


def _get_random_word() -> str:

    """
    Get a random word from the database

    Returns:
        The word retrieved from the database.

    Raises:
        DatabaseError: if no word could be retrieved from the database.
    """

    with db_cursor() as cur:
        cur.execute(Query.GetRandomWord)
        result = cur.fetchone()

    if result is None:
        raise DatabaseError("No se ha podido generar una palabra para el juego de hoy")

    return result["word"]


def _inspect_attempt(seed: str, word: str) -> List[CharInspection]:

    """
    Inspect every char in the given attempt against a seed word.

    Args:
        seed: The seed word, which the player tries to guess.
        word: The word given by the player.

    Returns:
        A list of `CharInspection` instances.

    Raises:
        GameError: if the seed word or the one given by the player has a wrong length.
    """

    if len(seed) != WORD_LENGTH:
        raise GameError(f"La palabra semilla tiene que tener [bold]{WORD_LENGTH}[/] letras")

    if len(word) != WORD_LENGTH:
        raise GameError(f"La palabra tiene que tener [bold]{WORD_LENGTH}[/] letras")

    result = []
    word = _normalize_word(word)
    upper_seed = seed.upper()
    occurrences = Counter(char[0] for char in upper_seed)

    for i in range(WORD_LENGTH):

        inspection = CharInspection(
            char=word[i].upper(),
            position=i,
        )

        if inspection.char in upper_seed and occurrences[inspection.char] > 0:
            if upper_seed[i] == inspection.char:
                inspection.valid = True
            else:
                inspection.missplaced = True
            occurrences[inspection.char] -= 1

        result.append(inspection)

    return result


def word_exists(word: str) -> bool:

    """
    Check if the given word exists in the database dictionary.

    Args:
        word: The word to check.

    Returns:
        True if the word exists in the database dictionary; otherwise False.
    """

    with db_cursor() as cur:
        cur.execute(Query.WordExists, (_normalize_word(word),))
        result = cur.fetchone()

    return result is not None and bool(result["total"])


def get_current_game() -> Game:

    """
    Get the current game if it's already ongoing; otherwise create a new one.

    Returns:
        An instance of `Game` with the current game.
    """

    today = datetime.now().strftime("%Y%m%d")

    # Retrieve the current game
    with db_cursor() as cur:
        cur.execute(Query.GetGameByDate, (today,))
        game = cur.fetchone()

    # Create a new game
    if game is None:
        word = _get_random_word()
        with db_cursor() as cur:
            cur.execute(Query.CreateGame, (word, today))
            game_id = cur.lastrowid
        return Game(game_id, word, today)

    # Retrieve the current game's attempts
    with db_cursor() as cur:
        cur.execute(Query.GetAttemptsByGame, (game["id"],))
        attempts = cur.fetchall()

    return Game(
        game["id"],
        game["word"],
        game["date"],
        [
            Attempt(
                attempt["id"],
                attempt["word"],
                bool(attempt["match"]),
                _inspect_attempt(
                    game["word"],
                    attempt["word"],
                ),
            )
            for attempt in attempts
        ],
    )


def play_game(game: Game, word: str) -> Attempt:

    """
    Try to guess the seed word.

    Args:
        game: The current game.
        word: The word given by the player.

    Returns:
        An instance of `Attempt` with the result.
    """

    inspection = _inspect_attempt(game.word, word)
    match = all([char.valid for char in inspection])

    # Create a new attempt
    with db_cursor() as cur:
        cur.execute(Query.CreateAttempt, (game.id, word, match))
        attempt_id = cur.lastrowid

    # Add the attempt to the current game
    return Attempt(
        id=attempt_id,
        word=word,
        match=match,
        chars=inspection,
    )



from enum import Enum, unique


@unique
class ErrorCode(str, Enum):

    """
    The available error codes.
    """

    DatabaseErrorCode = "palabros.databaseError"
    GameErrorCode = "palabros.gameError"


class PalabrosError(Exception):

    """
    Generic error.
    """

    @property
    def code(self) -> str:

        """
        Get the error code.

        Returns:
            The error code.
        """

        return self._code

    @property
    def message(self) -> str:

        """
        Get the error message.

        Returns:
            The error message.
        """

        return self._message

    def __init__(self, code: str, message: str) -> None:

        """
        Class constructor.

        Args:
            code: The error code.
            message: The message code.
        """

        self._code = code
        self._message = message

    def __str__(self) -> str:

        """
        Get a readable representation of the exception.

        Returns:
            A string representing the exception.
        """

        return self.message

    def __repr__(self) -> str:

        """
        Get an unambiguous representation of the exception.

        Returns:
            A string representing the exception.
        """

        return f"<{self.__class__.__name__} code={self.code}> {self.message}"


class DatabaseError(PalabrosError):

    """
    An exception raised when an error occurred while working with the database.
    """

    def __init__(self, message: str) -> None:

        """
        Class constructor.

        Args:
            message: The message code.
        """

        super().__init__(ErrorCode.DatabaseErrorCode, message)


class GameError(PalabrosError):

    """
    An exception raised when an error occurred while working a `Game` business logic.
    """

    def __init__(self, message: str) -> None:

        """
        Class constructor.

        Args:
            message: The message code.
        """

        super().__init__(ErrorCode.GameErrorCode, message)



class Query:

    """
    This is just a query catalog.
    """

    WordExists = """
        SELECT
            COUNT(*) AS total
        FROM
            words AS w
        WHERE
            w.word = ?;
    """

    GetRandomWord = """
        SELECT
            w.word
        FROM (
            SELECT
                w.word
            FROM
                words AS w
                LEFT OUTER JOIN
                    games AS g ON w.word = g.word
            WHERE
                g.id IS NULL
            ORDER BY
                w.frequency DESC
            LIMIT
                50
        ) as w
        ORDER BY
            RANDOM()
        LIMIT
            1;
    """

    GetGameByDate = """
        SELECT
            g.id,
            g.word,
            g.date
        FROM
            games AS g
        WHERE
            g.date = ?;
    """

    CreateGame = """
        INSERT INTO
            games (
                word,
                date
            )
        VALUES (
            ?,
            ?
        );
    """

    GetAttemptsByGame = """
        SELECT
            a.id,
            a.word,
            a.match
        FROM
            attempts AS a
        WHERE
            a.game_id = ?;
    """

    CreateAttempt = """
        INSERT INTO
            attempts (
                game_id,
                word,
                match
            )
        VALUES (
            ?,
            ?,
            ?
        );
    """

    TableExists = """
        SELECT
            COUNT(*) AS total
        FROM
            sqlite_master
        WHERE
            type = 'table'
        AND
            name = ?;
    """

    CreateVersionsTable = """
        CREATE TABLE IF NOT EXISTS versions (
            version TEXT PRIMARY KEY
        );
    """

    CreateWordsTable = """
        CREATE TABLE IF NOT EXISTS words (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            word TEXT NOT NULL UNIQUE,
            frequency REAL NOT NULL
        );
    """

    CreateGamesTable = """
        CREATE TABLE IF NOT EXISTS games (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            word TEXT NOT NULL,
            date TEXT NOT NULL,
            FOREIGN KEY (word) REFERENCES words (word)
        );
    """

    CreateAttemptsTable = """
        CREATE TABLE IF NOT EXISTS attempts (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            game_id INTEGER NOT NULL,
            word TEXT NOT NULL,
            match INTEGER NOT NULL,
            FOREIGN KEY (game_id) REFERENCES games (id)
        );
    """

    CreateGamesDateIndex = """
        CREATE UNIQUE INDEX IF NOT EXISTS idx_games_date ON games (date);
    """

    PopulateWordsTable = """
        INSERT OR IGNORE INTO words (word, frequency) VALUES {};
    """

    CreateVersion = """
        INSERT INTO
            versions (
                version
            )
        VALUES (
            ?
        );
    """



import typer

from palabros import business, console, database
from palabros.errors import DatabaseError, GameError

database.init()
app = typer.Typer()


@app.command("play")
def _play_game(word: str) -> None:

    """
    Try to guess the seed word using WORD.
    \f

    Args:
        word: The word which he player is playing with.

    Raises:
        Exit: if an error occurred while executing the command.
    """

    if len(word) != business.WORD_LENGTH:
        console.print_error(f"La palabra tiene que tener [bold]{business.WORD_LENGTH}[/] letras")
        raise typer.Exit()

    if not business.word_exists(word):
        console.print_error(f"La palabra [bold]{word}[/] no existe en el diccionario")
        raise typer.Exit()

    try:
        game = business.get_current_game()

        if not game.any_match() and game.any_attempt_left():
            attempt = business.play_game(game, word)
            game.attempts.append(attempt)

    except (DatabaseError, GameError) as err:
        console.print_error(err.message)
    except Exception as err:
        console.print_error(f"Error no controlado: {str(err)}")
    else:
        console.print_result(game)


@app.command("results")
def _view_results(date: str) -> None:

    """
    Check the results of a given DATE.
    \f

    Args:
        date: The date, given in YYYYMMDD format.

    Raises:
        Exit: if an error occurred while executing the command.
    """

    console.print_error("Esta funcionalidad aún no está implementada")
    raise typer.Exit()


if __name__ == "__main__":
    app()



from datetime import datetime, time, timedelta
from typing import List

from rich.console import Console
from rich.padding import Padding
from rich.table import Table, box

from palabros.business import WORD_LENGTH
from palabros.schemas import Attempt, Game


def _generate_attempt_row(attempt: Attempt) -> List[Padding]:

    """
    Generate a table row with the given attempt info.

    Args:
        attempt: The attempt result to show.

    Returns:
        A list of `rich.padding.Padding` instances.
    """

    return [
        Padding(
            char.char,
            1,
            style="bold black on green"
            if char.valid
            else "bold black on yellow"
            if char.missplaced
            else "bold white on black",
        )
        for char in sorted(attempt.chars, key=lambda x: x.position)
    ]


def _generate_empty_row(cols: int = WORD_LENGTH) -> List[str]:

    """
    Generate and empty table row.

    Args:
        cols: The number of columns.

    Returns:
        A list of `rich.padding.Padding` instances.
    """

    return [Padding("", 1) for _ in range(cols)]


def _create_table(game: Game) -> Table:

    """
    Create a table to show the given game status.

    Args:
        game: The game to show.

    Returns:
        An instance of `rich.table.Table`.
    """

    table = Table(
        show_header=False,
        show_footer=False,
        show_lines=True,
        box=box.HEAVY,
    )

    table.add_column(width=5, justify="center")
    table.add_column(width=5, justify="center")
    table.add_column(width=5, justify="center")
    table.add_column(width=5, justify="center")
    table.add_column(width=5, justify="center")

    for attempt in game.attempts:
        table.add_row(*_generate_attempt_row(attempt))

    for _ in range(game.get_attempts_left()):
        table.add_row(*_generate_empty_row())

    return table


def _get_countdown() -> str:

    """
    Calculate the countdown to the moment the user can play again.

    Returns:
        The countdown, given as a string.
    """

    now = datetime.now()
    tomorrow = now + timedelta(days=1)
    midnight = datetime.combine(tomorrow.date(), time(0, 0))
    countdown = midnight - now
    hours, remainder = divmod(countdown.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return f'{str(hours).rjust(2, "0")}:{str(minutes).rjust(2, "0")}:{str(seconds).rjust(2, "0")}'


def print_error(message: str) -> None:

    """
    Print the given error message in the console.

    Args:
        message: The message to print.
    """

    console = Console()
    console.print(f"\n{message}\n")


def print_result(game: Game) -> None:

    """
    Print the status of the given game in the console.

    Args:
        game: The game which status should be printed.
    """

    message = (
        f"¡Bien hecho! Siguiente palabra en [bold]{_get_countdown()}[/]."
        if game.any_match()
        else f"¡Has fallado! Te quedan [bold]{game.get_attempts_left()}[/] intentos."
        if game.any_attempt_left()
        else f"¡Tenías que acertar [bold]{game.word.upper()}[/]! Siguiente palabra en [bold]{_get_countdown()}[/]."
    )

    console = Console()
    console.print(_create_table(game))
    console.print(f"\n{message}\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['palabros']

package_data = \
{'': ['*']}

install_requires = \
['rich>=11.1.0,<12.0.0', 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['palabros = palabros.main:app']}

setup_kwargs = {
    'name': 'palabros',
    'version': '0.0.6',
    'description': 'Este CLI no es más que clon de Wordle en español',
    'long_description': '# palabros\n\nEsto no es más que un CLI inspirado en [Wordle ES](https://wordle.danielfrg.com/), que a su vez es una adaptación de [Wordle](https://www.powerlanguage.co.uk/wordle/). No hay mucho de original con **palabros**, más allá de lo _sexy friki_ de jugar a través de la terminal.\n\n![palabros screenshot](https://rawcdn.githack.com/vermicida/palabros/4ab0e73a58113195ca8c4e71698ceb832388f01d/assets/screenshot.png)\n\nPuedes leer el artículo [Creando un Wordle para consola](https://medium.com/@vermicida/creando-un-wordle-para-consola-b284a6ad10a0) de Medium para conocer con algo más de detalle cómo planteé el desarrollo de **palabros**.\n\n## Cómo jugar\n\n**palabros** está publicado en [PyPI](https://pypi.org/), por lo que su instalación es tan sencilla como:\n\n```bash\npip install palabros\n```\n\nSe muestra la instalación con [pip](https://pip.pypa.io/en/stable/cli/pip_install/), pero se puede usar cualquier otro gestor de paquetes, como [Poetry](https://python-poetry.org/docs/cli/#add), por ejemplo.\n\nUna vez instalado, tendremos disponible el comando `palabros play`, que podremos usar para jugar con una palabra de nuestra conveniencia:\n\n```bash\npalabros play coche\n```\n\nAl ejecutar el comando pueden darse las siguientes opciones:\n\n- Si es la primera ejecución del día, cuenta como primer intento de adivinar la palabra semilla\n- Si no es la primera ejecución del día, cuenta como un intento sucesivo de adivinar la palabra semilla -hasta un total de 6 intentos-\n- Si la palabra semilla se adivinó, o bien no se adivinó pero se consumió el número de intentos, se mostrará una cuenta atrás a cumplir para que se genere una nueva palabra semilla y, por tanto, se pueda volver a jugar\n\nComo resumen: puedes ejecutar `palabros play` cuantas veces quieras, pero solo las 6 primeras del día cuentan como intentos de adivinar la palabra semilla.\n\nSi le das al coco lo suficiente y lo combinas con un poco de suerte, darás con la palabra semilla :-)\n\n## Cómo trabajar\n\nSi quieres trastear con el código, hacer añadidos o cambios, necesitas saber que **palabros** está desarrollado con [Python 3.8](https://www.python.org/downloads/) y utiliza [Poetry](https://python-poetry.org/) como gestor de paquetes. Una vez tengas ambos instalados y, también, el repositorio clonado, debes hacer lo siguiente:\n\n```bash\npoetry install\n```\n\nEsto creará un nuevo entorno virtual para el proyecto e instalará en él todas las dependencias necesarias. Con esto ya podrás ejecutar **palabros** localmente y desarrollar a tu antojo.\n\n¡Ah, importante! Se hace uso de [pre-commit](https://pre-commit.com/) para validar automáticamente la calidad del código que se _commitea_, ejecutando por nosotros los checks de [pyupgrade](https://github.com/asottile/pyupgrade), [isort](https://github.com/PyCQA/isort), [black](https://github.com/psf/black) y [flake8](https://github.com/PyCQA/flake8). Antes de hacer tu primer commit no olvides configurar el hook:\n\n```bash\npoetry run pre-commit install\n```\n\nEsto es necesario hacerlo una sola vez tras clonar el repositorio e instalar las dependencias. Ahora, con cada `git commit`, tendremos una validación de calidad del código _by the face_ sin tener que intervenir manualmente.\n\n## Por hacer\n\nAunque **palabros** sea fruto de un par de tardes ociosas y no tenga pretensión alguna, sí que tiene cosas que son mejorables y, conforme el tiempo me lo vaya permitiendo, las iré atendiendo. Tengo fichadas las siguientes:\n\n- ~~El diccionario de palabras que usa quizá no sea el mejor. Muchas de ellas no son de uso común y, a su vez, no contiene otras que sí lo son; me gustaría encontrar un dataset o corpus de sustantivos que tenga información de frecuencia de uso. No me culpes mucho si te salen palabras que desconoces -a mí también me pasa-.~~ He reconstruido el diccionario desde cero usando 3.750 palabras del [Corpus de Referencia del Español Actual (CREA) - Listado de frecuencias](https://corpus.rae.es/lfrecuencias.html) que ofrece gratuitamente la RAE. Ha sido tedioso, ya que la fuente es un _.txt_ sin formato alguno y con el encoding incorrecto, pero ya está hecho. Gracias [@iknite](https://github.com/iknite) por la ayuda.\n- A nivel arquitectura hay una separación de capas evidente, pero no es perfecta. Sería genial marcar bien los límites y que cada módulo, clase y método asuma solo y exclusivamente las responsabilidades que son de su competencia.\n- Muchos pequeños detalles no están contemplados. Por ejemplo, la limpieza de las palabras que indica el usuario se limita a tildes y diéresis, pero se pueden colar otros gazapos.\n- La gestión de errores es algo -muy- parca.\n- Los tests unitarios. La ausencia de ellos, vamos.\n\n## Créditos\n\nAquí quiero mencionar a varias personas.\n\nPor un lado, la idea de **palabros** surje de [Wordle ES](https://wordle.danielfrg.com/), y ésta a su vez de [Wordle](https://www.powerlanguage.co.uk/wordle/). Mi trabajo se resume en poner un par de neuronas a codificar una versión CLI que nada tiene que ver oficialmente con los servicios mencionados. Si os mola **palabros**, dad las gracias a [Daniel Rodriguez](https://twitter.com/danielfrg) y a [Josh Wardle](https://twitter.com/powerlanguish), que son, respectivamente, los responsables de Wordle ES y Wordle.\n\nPor otro lado, agradecer a [Carlos Fenollosa](https://twitter.com/cfenollosa) el trabajo realizado para crear el [Diccionario libre en español](https://cfenollosa.com/blog/diccionario-libre-en-espanol---free-spanish-dictionary.html) que he utilizado para generar el diccionario de **palabros**. Todas las palabras están debidamente tipificadas en el diccionario: yo las he filtrado para quedarme exclusivamente con nombres (comunes, compuestos, femeninos, masculinos y ambiguos) de 5 letras. Suman más de 3600 palabras: nos da para jugar diaramiente a **palabros** los próximos 10 años.\n\n## Licencia\n\nLa licencia que he contemplado es [MIT](./LICENSE), que si no me equivoco, permite hacer con este código cualquier modificación, publicación y explotación posible. Dicho de otro modo, más directo: haced con este código lo que más os mole. Eso sí, si os hacéis millonarios con él, que lo dudo, por favor, invitadme a un café; es lo único que pido.\n',
    'author': 'Diego Herrera',
    'author_email': 'vermicida@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vermicida/palabros',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
